#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <stddef.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>

static char const service[]="login";

int write_buf(int const fd, char const* buf, size_t len) {
  while (len!=0) {
    ssize_t const nwrit=write(fd, buf, len);
    if (nwrit<0) {
      if (errno==EINTR) continue;
      return -1;
    }
    buf+=nwrit;
    len-=nwrit;
  }
  return 0;
}

int read_line(int const fd, char* buf, size_t size, size_t* const plen) {
  size_t len=0;
  while (size!=0) {
    ssize_t const nread=read(fd, buf, 1);
    if (nread<0) {
      if (errno==EINTR) continue;
      return -1;
    }
    if (nread==0)
      break;
    --size;
    ++len;
    if (*(buf++)=='\n') break;
  }
  *plen=len;
  return 0;
}

void check(char* const buf, size_t const len, size_t const max,
          char const* const name) {
  char const* err=null;
  if (len==0 || buf[len-1]!='\n')
    die4x(111, service, fatal, "EOF while reading ", name);
  if (len==max)
    err=" too long\n";
  else if (byte_chr(buf, len, '\0')!=len)
    err=" contains null characters\n";
  buf[len-1]='\0';
  if (err==null) return;
  buffer_puts(buffer_1, name);
  buffer_puts(buffer_1, err);
  buffer_flush(buffer_1);
  buffer_puts(buffer_2, service);
  buffer_puts(buffer_2, fatal);
  buffer_puts(buffer_2, name);
  buffer_puts(buffer_2, err);
  buffer_flush(buffer_2);
  _exit(111);
}

static char const var_cvm_plain[]="$CVM_PLAIN";

int main(int argc, char** argv) {
  char const* x;
  typedef char buf[512];
  static size_t const bufmax=(sizeof (buf))-1;
  buf ubuf, pbuf;
  size_t ulen, plen;
  struct termios normal, noecho;
#ifdef TCSASOFT
  int const tcsa=TCSAFLUSH|TCSASOFT;
#else
  int const tcsa=TCSAFLUSH;
#endif
  int ret, err;
  char const* const cvm=env_get(1+var_cvm_plain);
  struct stralloc sa={ 0 };

  if (cvm==null)
    die4x(111, service, fatal, var_cvm_plain, err_notset);
  if (tcgetattr(fileno_stdin, &normal)!=0)
    die5sys(111, service, fatal, err_unable, err_tcgetattr, err_);
  noecho=normal;
  noecho.c_lflag&=~(ECHO|ISIG);

  warn3x(service, "tty ", env_get("TTY"));

  x=env_get("PROMPT_LOGIN");
  if (x==null) x="login: ";
  if (write_buf(fileno_stdout, x, str_len(x))!=0)
    die5sys(111, service, fatal, err_unable, "write login prompt", err_);
  if (read_line(fileno_stdin, ubuf, bufmax, &ulen)!=0)
    die5sys(111, service, fatal, err_unable, "read username", err_);
  check(ubuf, ulen, bufmax, "username");

  warn3x(service, "attempt ", ubuf);

  if (env_get("DENY")!=null) {
    if (stralloc_copys(&sa, "ALLOW_")==0 ||
        stralloc_cats(&sa, ubuf)==0 ||
        stralloc_append(&sa, "")==0)
      die5sys(111, service, fatal, err_unable, err_alloc, err_);
    err=(env_get(sa.s)==null);
  } else {
    if (stralloc_copys(&sa, "DENY_")==0 ||
        stralloc_cats(&sa, ubuf)==0 ||
        stralloc_append(&sa, "")==0)
      die5sys(111, service, fatal, err_unable, err_alloc, err_);
    err=(env_get(sa.s)!=null);
  }
  if (err) {
    x="login denied";
    buffer_puts(buffer_1, x);
    buffer_puts(buffer_1, "\n");
    buffer_flush(buffer_1);
    die2x(100, service, x);
  }

  x=env_get("PROMPT_PASSWORD");
  if (x==null) x="Password: ";
  if (write_buf(fileno_stdout, x, str_len(x))!=0)
    die5sys(111, service, fatal, err_unable, "write password prompt", err_);
  if (tcsetattr(fileno_stdin, tcsa, &noecho)!=0)
    die5sys(111, service, fatal, err_unable, err_tcsetattr, err_);
  ret=read_line(fileno_stdin, pbuf, bufmax, &plen);
  err=errno;
  if (tcsetattr(fileno_stdin, tcsa, &normal)!=0)
    die5sys(111, service, fatal, err_unable, err_tcsetattr, err_);
  if (ret!=0) {
    errno=err;
    die5sys(111, service, fatal, err_unable, "read password", err_);
  }
  if (write_buf(fileno_stdout, "\n", 1)!=0)
    die5sys(111, service, fatal, err_unable, "write newline", err_);
  check(pbuf, plen, bufmax, "password");

  {
    char const* credentials[]={ ubuf, pbuf, null };
    ret=authenticate(cvm, credentials);
  }
  if (ret!=0) {
    if (ret==CVME_PERMFAIL) {
      x="login incorrect";
      buffer_puts(buffer_1, x);
      buffer_puts(buffer_1, "\n");
      buffer_flush(buffer_1);
      die2x(100, service, x);
    }
    switch (ret) {
      case CVME_GENERAL:     x="unspecified error";        break;
      case CVME_BAD_CLIDATA: x="invalid data from client"; break;
      case CVME_BAD_MODDATA: x="invalid data from module"; break;
      case CVME_IO:          x="I/O error";                break;
      default:               x="unknown error";            break;
    }
    buffer_puts(buffer_1, "error authenticating\n");
    buffer_flush(buffer_1);
    die3x(111, service, fatal, x);
  }

  warn2x(service, "login ok");

  {
    pid_t const pid=fork();
    switch (pid) {
      case -1: die5sys(111, service, fatal, err_unable, err_fork, err_);
      case 0: break;
      default:
        if (waitpid(pid, null, 0)<0)
          warn5sys(service, warning, err_unable, err_wait, err_);
        warn2x(service, "logout");
        _exit(0);
    }
  }

  if (pathexec_env(1+var_user, fact_username)==0)
    die6sys(111, service, fatal, err_unable, err_setvar, var_user, err_);
  if (stralloc_copyb(&sa, "", 0)==0)
    die5sys(111, service, fatal, err_unable, err_alloc, err_);
  {
    char const* c;
    if (fact_str(FACT_USERID, &c)!=0)
      if (pathexec_env(1+var_uid, c)==0)
        die6sys(111, service, fatal, err_unable, err_setvar, var_uid, err_);
    if (fact_str(FACT_GROUPID, &c)!=0) {
      if (pathexec_env(1+var_gid, c)==0)
        die6sys(111, service, fatal, err_unable, err_setvar, var_gid, err_);
      if (env_get("INCLUDE_GID")!=null)
        if (stralloc_copys(&sa, " ")==0 ||
            stralloc_cats(&sa, c)==0)
          die5sys(111, service, fatal, err_unable, err_alloc, err_);
    }
    if (fact_str(FACT_SUPP_GROUPID, &c)!=0)
      if (stralloc_cats(&sa, " ")==0 ||
          stralloc_cats(&sa, c)==0)
        die5sys(111, service, fatal, err_unable, err_alloc, err_);
  }
  x=env_get("EXTRA_GROUPS");
  if (x!=null && stralloc_cats(&sa, x)==0)
    die5sys(111, service, fatal, err_unable, err_alloc, err_);
  if (stralloc_append(&sa, "")==0)
    die5sys(111, service, fatal, err_unable, err_alloc, err_);
  if (pathexec_env(1+var_groups, sa.s)==0)
    die6sys(111, service, fatal, err_unable, err_setvar, var_groups, err_);
  if (pathexec_env(1+var_name, fact_realname)==0)
    die6sys(111, service, fatal, err_unable, err_setvar, var_name, err_);
  if (pathexec_env(1+var_home, fact_directory)==0)
    die6sys(111, service, fatal, err_unable, err_setvar, var_home, err_);
  if (pathexec_env(1+var_shell, fact_shell)==0)
    die6sys(111, service, fatal, err_unable, err_setvar, var_shell, err_);

  close(fileno_stderr);
  if (dup(fileno_stdout)!=fileno_stderr)
    _exit(111);

  {
    char* usercmd[]={ (char*)PACKAGE_HOME "/command/setstate",
                      (char*)"lofgGuh-", null };
    if (argc>1)
      ++argv;
    else
      argv=usercmd;
    pathexec(argv);
    die6sys(111, service, fatal, err_unable, err_exec, argv[0], err_);
  }
}
