#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <unistd.h>
#include <grp.h>
#include <signal.h>

extern int setgroups(size_t size, const gid_t *list);

static char const prog[]="setstate";

int main(int argc, char** argv) {
  char const* x;
  unsigned int i;
  unsigned long l;
  int do_chroot=0;
  int do_leader=0;
  int do_noleader=0;
  int do_setsid=0;
  int do_tty_chown=0;
  int do_tty_fg=0;
  int do_setgid=0;
  int do_setgroups=0;
  int do_setuid=0;
  int do_home=0;
  int do_dash=0;

  if (argc>1) {
    x=argv[1];
    ++argv;
    --argc;
  } else
    x="";

  while (*x!='\0') {
    switch (*x) {
      case 'r': do_chroot=1;    break;
      case 'l': do_leader=1;    break;
      case 'L': do_noleader=1;  break;
      case 's': do_setsid=1;    break;
      case 'o': do_tty_chown=1; break;
      case 'f': do_tty_fg=1;    break;
      case 'g': do_setgid=1;    break;
      case 'G': do_setgroups=1; break;
      case 'u': do_setuid=1;    break;
      case 'h': do_home=1;      break;
      case '-': do_dash=1;      break;
      default: {
        char opt[2]={ 0, 0 };
        *opt=*x;
        die5x(111, prog, fatal, "unrecognized option", err_, opt);
      }
    }
    ++x;
  }

  if (do_chroot!=0) {
    x=env_get(1+var_root);
    if (x==0)
      die4x(111, prog, fatal, var_root, err_notset);
    if (chdir(x)!=0)
      die6sys(111, prog, fatal, err_unable, err_chdirto, x, err_);
    if (chroot(".")!=0)
      die6sys(111, prog, fatal, err_unable, err_chrootto, x, err_);
  }

  if (do_leader!=0)
    if (setpgid(0, 0)!=0)
      die5sys(111, prog, fatal, err_unable, err_pgid, err_);

  if (do_noleader!=0) {
    pid_t pid;
    int pipefd[2];
    if (pipe(pipefd)==-1)
      die6sys(111, prog, fatal, err_unable, err_create, err_pipe, err_);
    pid=fork();
    if (pid==-1)
      die5sys(111, prog, fatal, err_unable, err_fork, err_);
    if (pid==0) {
      close(pipefd[0]);
      if (setpgid(0, 0)!=0)
        die6sys(111, prog, fatal, err_unable, err_pgid, " in child", err_);
      close(pipefd[1]);
      pause();
      _exit(0);
    }
    close(pipefd[1]);
    read(pipefd[0], &pipefd[1], 1);
    close(pipefd[0]);
    if (setpgid(getpid(), pid)!=0)
      die5sys(111, prog, fatal, err_unable, err_pgid, err_);
    kill(pid, SIGKILL);
    waitpid(pid, 0, 0);
  }

  if (do_setsid!=0)
    if (setsid()<0)
      die5sys(111, prog, fatal, err_unable, err_sid, err_);

  if (do_tty_fg!=0) {
    sigset_t ttou, old;
    if (sigemptyset(&ttou)!=0)
      die6sys(111, prog, fatal, err_unable, err_empty, err_sigset, err_);
    if (sigaddset(&ttou, SIGTTOU)!=0)
      die6sys(111, prog, fatal, err_unable, err_addto, err_sigset, err_);
    if (sigprocmask(SIG_BLOCK, &ttou, &old)!=0)
      die5sys(111, prog, fatal, err_unable, err_sigmask, err_);
    if (tcsetpgrp(0, getpgid(0)))
      die5sys(111, prog, fatal, err_unable, "set foreground process group id",
              err_);
    if (sigprocmask(SIG_SETMASK, &old, 0)!=0)
      die5sys(111, prog, fatal, err_unable, err_sigmask, err_);
  }

  if (do_tty_chown!=0) {
    x=env_get(1+var_uid);
    if (x==0)
      die4x(111, prog, fatal, var_uid, err_notset);
    i=scan_ulong(x, &l);
    if (i==0 || x[i]!='\0')
      die4x(111, prog, fatal, var_uid, err_malformed);
    if (l!=(uid_t)l)
      die5x(111, prog, fatal, var_uid, err_oflow, "uid_t");
    if (fchown(fileno_stdin, l, -1)!=0)
      die6sys(111, prog, fatal, err_unable, err_chown, "fd 0", err_);
  }

  if (do_setgid!=0) {
    x=env_get(1+var_gid);
    if (x==0)
      die4x(111, prog, fatal, var_gid, err_notset);
    i=scan_ulong(x, &l);
    if (i==0 || x[i]!='\0')
      die4x(111, prog, fatal, var_gid, err_malformed);
    if (l!=(gid_t)l)
      die5x(111, prog, fatal, var_gid, err_oflow, "gid_t");
    if (setgid(l)!=0)
      die6sys(111, prog, fatal, err_unable, "setgid to ", x, err_);
  }

  if (do_setgroups!=0) {
    size_t ngroups;
    gid_t gid, * groups;
    static char const toolong[]="too long";
    x=env_get(1+var_groups);
    if (x==0)
      die4x(111, prog, fatal, var_groups, err_notset);
    {
      char const* y=x;
      ngroups=0;
      while (*y==' ') {
        ++ngroups;
        if (ngroups==0)
          die4x(111, prog, fatal, var_groups, toolong);
        ++y;
        y+=str_chr(y, ' ');
      }
    }
    i=ngroups*sizeof (gid_t);
    if (i/sizeof (gid_t)!=ngroups)
      die4x(111, prog, fatal, var_groups, toolong);
    groups=(gid_t*)alloc(i);
    if (groups==0)
      die5sys(111, prog, fatal, err_unable, err_alloc, err_);
    ngroups=0;
    while (*x!='\0') {
      if (*x!=' ')
        die4x(111, prog, fatal, var_groups, err_malformed);
      ++x;
      i=scan_ulong(x, &l);
      x+=i;
      if (i==0)
        die4x(111, prog, fatal, var_groups, err_malformed);
      gid=l;
      if (gid!=l)
        die5x(111, prog, fatal, "group id", err_oflow, "gid_t");
      groups[ngroups]=gid;
      ++ngroups;
    }
    if (setgroups(ngroups, groups)!=0)
      die5sys(111, prog, fatal, err_unable, "set supplementary groups", err_);
  }

  if (do_setuid!=0) {
    x=env_get(1+var_uid);
    if (x==0)
      die4x(111, prog, fatal, var_uid, err_notset);
    i=scan_ulong(x, &l);
    if (i==0 || x[i]!='\0')
      die4x(111, prog, fatal, var_uid, err_malformed);
    if (l!=(uid_t)l)
      die5x(111, prog, fatal, var_uid , err_oflow, "uid_t");
    if (setuid(l)!=0)
      die6sys(111, prog, fatal, err_unable, "setuid to ", x, err_);
  }

  if (do_home!=0) {
    x=env_get(1+var_home);
    if (x==0)
      die4x(111, prog, fatal, var_home, err_notset);
    if (chdir(x)!=0)
      die6sys(111, prog, fatal, err_unable, err_chdirto, x, err_);
  }

  {
    char* args[]={ 0, 0 };
    stralloc sa={ 0 };
    if (argc>1) {
      argv++;
      x=argv[0];
    } else {
      x=env_get(1+var_shell);
      if (x==0)
        die4x(111, prog, fatal, var_shell, err_notset);
      argv=args;
      argv[0]=(char*)x;
    }
    if (do_dash!=0) {
      i=str_rchr(x, '/');
      if (x[i]=='\0')
        i=0;
      else
        ++i;
      if (stralloc_copys(&sa, "-")==0 ||
          stralloc_cats(&sa, x+i)==0 ||
          stralloc_append(&sa, "")==0)
        die5sys(111, prog, fatal, err_unable, err_alloc, err_);
      argv[0]=sa.s;
    }
    pathexec_run(x, argv, environ);
  }
  die6sys(111, prog, fatal, err_unable, err_exec, x, err_);
}
